##
# Copyright (c) 2010-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Toolkit for building a Data-Access Layer (DAL).

This includes an abstract representation of SQL objects like tables, columns,
sequences and queries, a parser to convert your schema to that representation,
and tools for working with it.

In some ways this is similar to the low levels of something like SQLAlchemy, but
it is designed to be more introspectable, to allow for features like automatic
caching and index detection.  NB: work in progress.
"""
