##
# Copyright (c) 2014-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
LDAP directory service.
"""

__all__ = [
    "LDAPError",
    "LDAPConfigurationError",
    "LDAPConnectionError",
    "LDAPBindAuthError",
    "RecordTypeSchema",
    "DirectoryService",
    "LDAPAttribute",
    "LDAPObjectClass",
    "FieldName",
]


from ._service import (
    LDAPError,
    LDAPConfigurationError,
    LDAPConnectionError,
    LDAPBindAuthError,
    RecordTypeSchema,
    DirectoryService,
    FieldName
)
from ._constants import (
    LDAPAttribute,
    LDAPObjectClass,
)
